function sndcft_vot

tracker_path = fileparts(mfilename('fullpath'));
addpath(fullfile(tracker_path, 'vot'));

% *************************************************************
% VOT: Set random seed to a different value every time.
% *************************************************************
try
    % Simple check for Octave environment
    OCTAVE_VERSION;
    rand('seed', sum(clock));
    pkg load image;
catch
    RandStream.setGlobalStream(RandStream('mt19937ar', 'Seed', sum(clock)));
end

% **********************************
% VOT: Get initialization data
% **********************************
[handle, image, region] = vot('rectangle', 'rgbt');

%%%%% 
% Initialize the tracker
tracker_path = fileparts(mfilename('fullpath'));
addpath(fullfile(tracker_path, 'external_libs/fhog'));
addpath(fullfile(tracker_path, 'external_libs/mtimesx'));
addpath(fullfile(tracker_path, 'external_libs/mexResize'));
addpath(fullfile(tracker_path, 'external_libs/cvNN'));
addpath(fullfile(tracker_path, 'features'));

%%%%% Setting parameters for local use.
search_area_scale    = 4.25;     % 4.25
search_area_shape    = 'square'; % the shape of the training/detection window:
                                 % 'proportional', 'square' or 'fix_padding'
output_sigma_factor  = 1/10;
learning_rate_filt   = 0.025;
filter_max_area      = 32^2;     % 32
win_param            = 1;
nScales              = 3;
scale_step           = 1.03;
reg_window_std       = 0.25;     % 0.25
ctx_amount           = 1.5;     % 1.5
channel_independent  = false;	% false

cnn_params.net_conf    = fullfile(tracker_path, 'nets/train_val.prototxt');
cnn_params.net_weights = fullfile(tracker_path, 'nets/squeezenet_v1.1.caffemodel');
cnn_params.scale       = 1;
cnn_params.size        = [137, 137]; % [177, 177] for 'fire4/concat'
cnn_params.mean        = [104, 117, 123, 0];
cnn_params.layer       = {'fire2/concat'}; %{25};
cnn_params.target      = 'cpu'; % 'cpu'

% Coordinates interpolation
newton_iterations    = 5;  % 50
interpolate_response = 4;  % correlation score interpolation strategy:
                           % 0 - off, 1 - feature grid, 2 - pixel grid, 4 - Newton's method

% Global feature parameters
cell_size             = 4;
hog_params.nDim       = 31;

features{1} = {
    struct('getFeature', @get_fhog, 'fparams', hog_params),...
    struct('getFeature', @get_cnn_layer_cv, 'fparams', cnn_params),...
};
features{2} = {
    struct('getFeature', @get_fhog, 'fparams', hog_params),...
    struct('getFeature', @get_cnn_layer_cv, 'fparams', cnn_params),...
};

% Visualization
visualize = false;

%%%%%
% Get a first frame
im = imread(image{1});
im_ir = imread(image{2});

% 
rect = round(region);
%display(rect)

pos = floor([rect(2) + rect(4) / 2, rect(1) + rect(3) / 2]);
target_sz = floor([rect(4), rect(3)]);
init_target_sz = target_sz;

% Set the feature ratio to the feature-cell size
featureRatio = cell_size;
search_area = prod(init_target_sz / featureRatio * search_area_scale);

global_feat_params.cell_size = cell_size;
%global_feat_params.cell_selection_thresh = cell_selection_thresh;

if (search_area > filter_max_area)
    currentScaleFactor = sqrt(search_area / filter_max_area);
else
    currentScaleFactor = 1.0;
end

% target size at the initial scale
base_target_sz = target_sz / currentScaleFactor;

% window size, taking padding into account
switch (search_area_shape)
    case 'proportional'
        % proportional area, same aspect ratio as the target
        sz = floor(base_target_sz * search_area_scale);     
    case 'square'
        % square area, ignores the target aspect ratio
        sz = repmat(sqrt(prod(base_target_sz * search_area_scale)), 1, 2); 
    case 'fix_padding'
        % const padding
        sz = base_target_sz + ...
            sqrt(prod(base_target_sz * search_area_scale) + ...
            (base_target_sz(1) - base_target_sz(2)) / 4) - sum(base_target_sz) / 2; 
    otherwise
        error(['Unknown "params.search_area_shape". Must be '...
               '''proportional'', ''square'' or ''fix_padding''']);
end

% set the size to exactly match the cell size
sz = round(sz / featureRatio) * featureRatio;

pixels = {get_pixels(im, pos, round(sz * currentScaleFactor), sz), ...
          get_pixels(im_ir, pos, round(sz * currentScaleFactor), sz)};
      
k = 1;
f = {};
for i = 1:length(pixels)
    f_ = get_features(pixels{i}, features{i}, global_feat_params);
    for j = 1:length(f_)
        f_r{k} = mean([size(pixels{i}, 1) / size(f_{j}, 1), ...
                       size(pixels{i}, 2) / size(f_{j}, 2)]); % !!!
        k = k + 1;
    end
    f((end + 1):(end + length(f_))) = f_;
end

use_sz = [max(cellfun(@(f) size(f, 1), f)), ...
          max(cellfun(@(f) size(f, 2), f))]; % !!!

% Construct the label function - correlation output, 2D gaussian function,
% with a peak located upon the target
R_ = cell(1, length(f));
for i = 1:length(f)
    s = [size(f{i}, 1), size(f{i}, 2)];
    
    out_sigma = sqrt(prod(max(floor(base_target_sz / f_r{i}), 2))) * output_sigma_factor;
    rg        = circshift(-floor((s(1) - 1) / 2):ceil((s(1) - 1) / 2), ...
                          [0 -floor((s(1) - 1) / 2)]);
    cg        = circshift(-floor((s(2) - 1) / 2):ceil((s(2) - 1) / 2), ...
                          [0 -floor((s(2) - 1) / 2)]);
    [rs, cs]  = ndgrid(rg, cg);
    r         = exp(-0.5 * (((rs .^ 2 + cs .^ 2) / out_sigma ^ 2)));
    % FFT of r
    R_{i}    = fft2(single(r));
end

if (interpolate_response == 1)
    interp_sz = use_sz * featureRatio;
else
    interp_sz = use_sz;
end

% Construct cosine window
cos_window = cell(1, length(f));
for i = 1:length(f)
    s = [size(f{i}, 1), size(f{i}, 2)];
    cos_window{i} = single(tukeywin(s(1), win_param) * ...
                           tukeywin(s(2), win_param)');
end

% Compute scales and weights
if (nScales > 0)
    scale_exp = (-floor((nScales-1)/2):ceil((nScales-1)/2));
    scaleFactors = scale_step .^ scale_exp;
    min_scale_factor = scale_step ^ ceil(log(max(5 ./ sz)) / log(scale_step));
    max_scale_factor = scale_step ^ floor(log(min([size(im,1) size(im,2)] ./ base_target_sz)) / log(scale_step));
    
    scaleWeights = exp(-0.5 * scale_exp .^ 2 / 16);
end

if (interpolate_response >= 3)
    % Pre-computes the grid that is used for socre optimization
    ky = circshift(-floor((use_sz(1) - 1)/2) : ceil((use_sz(1) - 1)/2), ...
                   [1, -floor((use_sz(1) - 1)/2)]);
    kx = circshift(-floor((use_sz(2) - 1)/2) : ceil((use_sz(2) - 1)/2), ...
                   [1, -floor((use_sz(2) - 1)/2)])';
end

frame = 1;
while true
    if (frame > 1)
        % **********************************
        % VOT: Get next frame
        % **********************************
    
        [handle, image] = handle.frame(handle);

        if (isempty(image))
            break;
        end;

        % Perform a tracking step, obtain new region
        im = imread(image{1});
        im_ir = imread(image{2});

        % do not estimate translation and scaling on the first frame,
        % since we just want to initialize the tracker there
        responsef = cell(nScales, length(f));
        for scale_ind = 1:nScales
            pixels = {get_pixels(im, pos, round(sz*currentScaleFactor*scaleFactors(scale_ind)), sz), ...
                      get_pixels(im_ir, pos, round(sz*currentScaleFactor*scaleFactors(scale_ind)), sz)};
                        
            f = {};
            for i = 1:length(pixels)
                f_ = get_features(pixels{i}, features{i}, global_feat_params);
                f((end + 1):(end + length(f_))) = f_;
            end
            
            for i = 1:length(f)
                F = fft2(bsxfun(@times, f{i}, cos_window{i}));
                corr = sum(F .* Gm{i}, 3);
                corr = corr * scaleWeights(scale_ind);
                responsef{scale_ind, i} = corr;
            end
        end
        
        % Interpolate the response so it has the same size as the image patch
        if (interpolate_response == 2)
            % use dynamic interp size
            interp_sz = floor(size(y) * featureRatio * currentScaleFactor);
        end
        responsef_padded = zeros(interp_sz(1), interp_sz(2), ...
                                 size(responsef, 2), size(responsef, 1));
        for scale_idx = 1:size(responsef, 1)
            for i = 1:size(responsef, 2)
                responsef_padded(:, :, i, scale_idx) = resizeDFT2(responsef{scale_idx, i}, interp_sz);
            end
        end
        responsef_padded = squeeze(sum(responsef_padded, 3));
        
        % response in the spatial domain
        response = ifft2(responsef_padded, 'symmetric');
        
        % find maximum peak
        if (interpolate_response == 3)
            error('Invalid parameter value for interpolate_response');
        elseif (interpolate_response == 4)
            [disp_row, disp_col, sind] = resp_newton(response, responsef_padded, newton_iterations, ky, kx, use_sz);
        else
            [row, col, sind] = ind2sub(size(response), find(response == max(response(:)), 1));
            disp_row = mod(row - 1 + floor((interp_sz(1)-1)/2), interp_sz(1)) - floor((interp_sz(1)-1)/2);
            disp_col = mod(col - 1 + floor((interp_sz(2)-1)/2), interp_sz(2)) - floor((interp_sz(2)-1)/2);
        end
        
        scale = scaleFactors(sind);
        % calculate translation
        switch (interpolate_response)
            case 0
                translation_vec = round([disp_row, disp_col] * featureRatio * currentScaleFactor * scale);
            case 1
                translation_vec = round([disp_row, disp_col] * currentScaleFactor * scale);
            case 2
                translation_vec = round([disp_row, disp_col] * scale);
            case 3
                translation_vec = round([disp_row, disp_col] * featureRatio * currentScaleFactor * scale);
            case 4
                translation_vec = round([disp_row, disp_col] * featureRatio * currentScaleFactor * scale);
        end
        
        % set the scale
        currentScaleFactor = currentScaleFactor * scale;
        % adjust to make sure we are not to large or to small
        if (currentScaleFactor < min_scale_factor)
            currentScaleFactor = min_scale_factor;
        elseif (currentScaleFactor > max_scale_factor)
            currentScaleFactor = max_scale_factor;
        end
        
        % update position
        pos = pos + translation_vec;
    end
    
    %%%%% Training
    % Extract training sample image region
    pixels = {get_pixels(im, pos, round(sz * currentScaleFactor), sz), ...
              get_pixels(im_ir, pos, round(sz * currentScaleFactor), sz)};
          
    % Extract features and do windowing
    f = {};
    for i = 1:length(pixels)
        f_ = get_features(pixels{i}, features{i}, global_feat_params);
        f((end + 1):(end + length(f_))) = f_;
    end
    
    F_ = cell(1, length(f));
    for i = 1:length(f)
        F_{i} = fft2(bsxfun(@times, f{i}, cos_window{i}));
    end
     
    %%%%%
    % ADMM
    assert(length(F_) == length(R_));
    if (frame == 1)
        G_ = cell(1, length(F_));
        U_ = cell(1, length(F_));
        w2_ = cell(1, length(F_));
        
        for i = 1:length(F_)
            initRho = 1;
            maxRho = 1000;
            numIt = 2;
            beta = 30;
            
            G_{i} = zeros(size(F_{i}), 'single');
            U_{i} = zeros(size(F_{i}), 'single');

            s = [size(F_{i}, 1), size(F_{i}, 2)];
            p = 100; %
            m = 0.01 * ones(s, 'single');
            tf_sz = round(ctx_amount * base_target_sz / f_r{i});
            tf_sz = min(tf_sz, s);
            m((1:tf_sz(1)) + floor((s(1) - tf_sz(1)) / 2), ...
              (1:tf_sz(2)) + floor((s(2) - tf_sz(2)) / 2)) = p;
            m = imfilter(m, fspecial('gaussian', [7, 7], reg_window_std)); %
            w = p - m;

            w2_{i} = 2 * w .^ 2;
        end
    end
    
    for i = 1:length(F_)
        F = F_{i};
        R = R_{i};
        w2 = w2_{i};
        G = G_{i};
        U = U_{i};
        %G = zeros(size(F), 'single');
        %U = zeros(size(F), 'single');
        
        rho = initRho;
        cFR = bsxfun(@times, conj(F), R);
        if (~channel_independent)
            scFF = sum(conj(F) .* F, 3);
        end
        for it = 1:numIt
            if (channel_independent)
                H = (cFR + (rho / 2) .* (G - U)) ./ (conj(F) .* F + (rho / 2));
            else
                t = (2 / rho) * cFR + (G - U);
                H = t - bsxfun(@times, conj(F), sum(F .* t, 3) ./ (scFF + (rho / 2)));
            end

            w = (rho / 2) ./ (w2 + (rho / 2));
            G = fft2(bsxfun(@times, real(ifft2(H + U)), w));

            U = U + (H - G);

            if (rho ~= maxRho)
                rho_ = rho;
                rho = min(rho * beta, maxRho);
                U = U .* (rho_ / rho);
            end
        end
        G_{i} = G;
        U_{i} = U;
    end
    
    %%%%% Update
    if (frame == 1)
        Gm = G_;
    else
        for i = 1:length(G_)
            Gm{i} = ((1 - learning_rate_filt) * Gm{i}) + (learning_rate_filt * G_{i});
        end
    end
    %%%%%
    
    target_sz = floor(base_target_sz * currentScaleFactor);
    
    % Position
    region = double([pos([2,1]) - target_sz([2,1])/2, target_sz([2,1])]);
    frame = frame + 1;
    
    % if (visualize)
    %     if (~exist('h_img', 'var'))
    %         h_img = imshow(im);
    %         h_rect = rectangle('Position', region);
    %     else
    %         set(h_img, 'CData', im);
    %         set(h_rect, 'Position', region);
    %     end        
    % end
    
    % **********************************
    % VOT: Report position for frame
    % **********************************
    handle = handle.report(handle, region);

end;

% **********************************
% VOT: Output the results
% **********************************
handle.quit(handle);

end
